/** 
 *******************************************************************************
 * @file malloc_utils.c
 * 
 * Copyright (c) 2010-2013 LG Electronics, Inc.
 * 
 * @brief Useful malloc utility functions.
 * 
 * @author Tom Hughes
 * 
 * @version  0.10  2010-08-09  <tdh>  Initial creation
 *******************************************************************************
 */

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/file.h>
#include <stdbool.h>

#define ERROR_BUF_SIZE_BYTES    64
static char err_buf[ERROR_BUF_SIZE_BYTES];

/** 
 *******************************************************************************
 * @brief Safe print routine that can be called from malloc without leading to
 * recursive malloc calls that might fail under ENOMEM conditions.
 * 
 * @param  s    IN      message to print 
 *******************************************************************************
 */
static void _malloc_print(const char* s, bool needs_lock)
{
    int bytes_written = 0;
    int ret = 0;
    int len = strlen(s);

    if (needs_lock) {
        flock(STDERR_FILENO, LOCK_EX);
    }

    while (bytes_written < len)
    {
        ret = write(STDERR_FILENO, s, len - bytes_written);

        if (ret < 0) {
            /* not much we can do in case of failure */
            goto done;
        }

        /* partial write, so we account for amount written */
        bytes_written += ret;
    }

done:
    if (needs_lock) {
        flock(STDERR_FILENO, LOCK_UN);
    }
}

void malloc_print(const char* s)
{
    _malloc_print(s, true);
}

void malloc_print_unlocked(const char* s)
{
    _malloc_print(s, false);
}

/** 
 *******************************************************************************
 * @brief perror replacement that doesn't allocate memory.
 * 
 * @param  s    IN      message to print
 *******************************************************************************
 */
void malloc_perror(const char* s)
{
    char *err_str = strerror_r(errno, err_buf, sizeof(err_buf));
    
    flock(STDERR_FILENO, LOCK_EX);
    malloc_print_unlocked(s);
    if (err_str) {
        malloc_print_unlocked(" : ");
        malloc_print_unlocked(err_str);
    }
    malloc_print_unlocked("\n");
    fsync(STDERR_FILENO);
    flock(STDERR_FILENO, LOCK_UN);
}
